/*
 * Copyright (c) 2010-2011 by inovat, innovative systeme - verkehr - tunnel - technik, Dipl.-Ing. H. C. Kniss
 *
 * This file is part of de.inovat.dua.pufferlzzs.PufferAbrufLzzs
 *
 * de.inovat.dua.pufferlzzs.PufferAbrufLzzs is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * de.inovat.dua.pufferlzzs.PufferAbrufLzzs is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * <p>
 * You should have received a copy of the GNU General Public License
 * along with de.inovat.dua.pufferlzzs.PufferAbrufLzzs; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Contact Information:
 * inovat, Dipl.-Ing. H. C. Kniss
 * Koelner Strasse 30
 * D-50859 Koeln
 * +49 (0)2234 4301 800
 * info@invat.de
 * www.inovat.de
 */



package de.inovat.dua.pufferlzzs;

//~ NICHT JDK IMPORTE =========================================================

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.application.StandardApplication;
import de.bsvrz.sys.funclib.application.StandardApplicationRunner;
import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList;
import de.bsvrz.sys.funclib.debug.Debug;

import java.text.SimpleDateFormat;
import java.util.*;

import static java.util.GregorianCalendar.MILLISECOND;
import static java.util.GregorianCalendar.MINUTE;
import static java.util.GregorianCalendar.SECOND;

//~ KLASSEN ===================================================================

/**
 * Klasse, die die den Pufferabfruf der TLS-De-Daten für die LZZS durchführt.
 *
 * @author inovat, innovative systeme - verkehr - tunnel - technik
 * @author Dipl.-Ing. Hans Christian Kniß (HCK)
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class PufferAbrufLzzs implements StandardApplication {

    /** WarteZeit in ms, bis die Auftragsliste erneut geprüft wird, nachdem keinerlei Aufträge mehr vorlagen * */
    private static final long WARTEZEIT_AUF_NAECHSTEN_AUFTRAG = 60 * 1000;  // 1 Minute

    /** DebugLogger fuer Debug-Ausgaben. */
    private static Debug debug;

    //~ FELDER ================================================================

    private long _bisZeitPunkt;

    /** Datenverteilerverbindung */
    private ClientDavInterface _dav;
    private int                _intervallDauerInStunden;
    private String             _intervallDauerInStundenAufrufParameter;

    /** Die Konfiguration des DaV. */
    private DataModel _konfiguration;
    private String    _listePidsAufrufParameter;

    /** Map der DeLve255 für Pufferabruf mit der Menge der zugeordneten FS */
    Map<SystemObject, Set<SystemObject>> _mapDe255ZuFs;

    /** Map aller DE-LVE und der zugeordneten DE255 der LVE */
    Map<SystemObject, SystemObject> _mapDeLveZuDe255;

    /** Map aller DE-LVE und der zugeordneten FS von LZZS */
    Map<SystemObject, SystemObject>  _mapDeLveZuFs;
    private int                      _minimaleWarteZeitInMinuten;
    private String                   _minimaleWarteZeitInMinutenAufrufParameter;
    private int                      _nachberechnungInStunden;
    private String                   _nachberechnungInStundenAufrufParameter;
    private long                     _vonZeitPunkt;
    private long                     _vonZeitPunktAufrufParameter;
    private List<String>             _listePidsLzzs               = new ArrayList<String>();
    private List<SystemObject>       _listeLzzsSystemObjekte      = new ArrayList<SystemObject>();
    private List<String>             _listePidsLzzsNichtVorhanden = new ArrayList<String>();
    private Map<String, VewDeLve255> _mapVewDeLve255              = new HashMap<String, VewDeLve255>();

    //~ METHODEN ==============================================================

    /**
     * Main-Methode der SWE PufferAbrufLzzs.
     *
     * @param args Aufrufargumente der Applikation.
     */
    public static void main(String[] args) {
        PufferAbrufLzzs pufferAbrufLzzs = new PufferAbrufLzzs();

        try {
            StandardApplicationRunner.run(pufferAbrufLzzs, args);

            // Debug erst hier initialisieren, da erst ab hier verfügbar !!
            debug = Debug.getLogger();
            pufferAbrufLzzs.auswertungAufrufParameterListePids();
            pufferAbrufLzzs.auswertungAufrufParameterVonBisIntervallNachberechnung();
            pufferAbrufLzzs.initialisierung();
            debug.config(pufferAbrufLzzs.ausgabeKonfigurationsZusammenfassung());
        }
        catch (RuntimeException e) {
            e.printStackTrace();
            System.exit(-1);
        }

        pufferAbrufLzzs.run();
    }

    /**
     * Gibt Informationen zur Konfiguration aus. Ist nur zu Anlaysezwecken bzw. zur zusammenfassenden Ausgabe der
     * Konfigurationsphase dieser Applikatione interessant.
     *
     * @return Informationen zur Konfiguration (Aufbau der Ausgabe kann sich ändern).
     */
    private String ausgabeKonfigurationsZusammenfassung() {
        StringBuffer sb = new StringBuffer();

        // Listen der SystemObjekte sortieren...
        Collections.sort(_listeLzzsSystemObjekte, new Comparator<SystemObject>() {
            public int compare(SystemObject o1, SystemObject o2) {
                return o1.getPid().compareTo(o2.getPid());
            }
        });
        Collections.sort(_listePidsLzzsNichtVorhanden);
        sb.append(String.format("%n%n"));
        sb.append(String.format("%n####################################################################################%n"));
        sb.append(String.format("START: Konfigurationszusammenfassung für PufferAbrufLzzs"));
        sb.append(String.format("%n####################################################################################%n"));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Im Aufrufparameter angegebene PIDs, zu denen keine SystemObjekte existieren"));
        sb.append(String.format("%n====================================================================================%n"));

        int i = 0;

        for (String pid : _listePidsLzzsNichtVorhanden) {
            sb.append(String.format("%04d: %s%n", ++i, pid));
        }

        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Liste der PIDs der verwendeten Systemobjekte"));
        sb.append(String.format("%n====================================================================================%n"));

        int j = 0;

        for (SystemObject so : _listeLzzsSystemObjekte) {
            sb.append(String.format("%04d: %s%n", ++j, so.getPidOrNameOrId()));
        }

        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Liste der DeLve255 zum Pufferabruf und den zugeordneten Fahrstreifenobjekten"));
        sb.append(String.format("%n====================================================================================%n"));

        int k = 0;

        for (SystemObject soDeLve255 : _mapDe255ZuFs.keySet()) {
            sb.append(String.format("%04d: %s%n", ++k, soDeLve255.getPidOrNameOrId()));

            int l = 0;

            for (SystemObject soFs : _mapDe255ZuFs.get(soDeLve255)) {
                sb.append(String.format("      %04d: %s%n", ++l, soFs.getPidOrNameOrId()));
            }
        }

        SimpleDateFormat datumsFormatierer = new SimpleDateFormat("dd.MM.yyyy HH:mm:ss,SSS");

        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Aufrufparameter -intervallDauerInStunden=..."));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Parameter beim Aufruf (Vorgabe)       : [%s]%n", _intervallDauerInStundenAufrufParameter));
        sb.append(String.format("Parameter nach Auswertung (Verwendung): [%s]%n", _intervallDauerInStunden));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Aufrufparameter -nachberechnungInStunden=..."));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Parameter beim Aufruf (Vorgabe)       : [%s]%n", _nachberechnungInStundenAufrufParameter));
        sb.append(String.format("Parameter nach Auswertung (Verwendung): [%s]%n", _nachberechnungInStunden));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Aufrufparameter -vonZeitPunkt=..."));
        sb.append(String.format("%n====================================================================================%n"));

        Date datum = new Date(_vonZeitPunktAufrufParameter);

        datum.toString();
        sb.append(String.format("Parameter beim Aufruf (Vorgabe)       : [%s]%n", datumsFormatierer.format(new Date(_vonZeitPunktAufrufParameter))));
        sb.append(String.format("Parameter nach Auswertung (Verwendung): [%s]%n", datumsFormatierer.format(new Date(_vonZeitPunkt))));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Aufrufparameter -minimaleWarteZeitInMinuten=..."));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Parameter beim Aufruf (Vorgabe)       : [%s]%n", _minimaleWarteZeitInMinutenAufrufParameter));
        sb.append(String.format("Parameter nach Auswertung (Verwendung): [%s]%n", _minimaleWarteZeitInMinuten));
        sb.append(String.format("%n====================================================================================%n"));
        sb.append(String.format("Auftragsliste initial..."));
        sb.append(String.format("%n====================================================================================%n"));

        int m = 0;

        for (Auftrag auftrag : AuftragsListe.getInstanz().holeKopieAuftragsListe()) {
            sb.append(String.format("%04d: %s%n", ++m, auftrag.toString()));
        }

        sb.append(String.format("%n%n"));
        sb.append(String.format("%n####################################################################################%n"));
        sb.append(String.format("ENDE: Konfigurationszusammenfassung für PufferAbrufLzzs"));
        sb.append(String.format("%n####################################################################################%n"));

        return sb.toString();
    }

    /**
     * Wertet die Aufrufparameter
     * <p>
     * -listePids
     * <p>
     * wie folgt aus:
     * <p>
     * Es werden alle Systemobjekte vom Typ <tt>typ.langZeitZählstelle</tt> ermittelt und dabei diejenigen für die
     * weiteren Berechnungen übernommen, deren Pid in der Liste der Pids aufgeführt ist (bzw. alle, falls Aufruf mit
     * <tt>-listePids=alle</tt> erfolgt). Wird kein passendes Objekt gefunden, wird das Programm mit einer Fehlermeldung
     * abgebrochen.
     */
    private void auswertungAufrufParameterListePids() {
        try {

            // Pids der zu berücksichtigenden Objekte ermitteln
            _listePidsLzzs = Arrays.asList(_listePidsAufrufParameter.split("(:|,|;)"));

            // Alle möglichen SystemObjekte ermittlen...
            List<SystemObject> listeAlleLzzsSystemObjekte = _konfiguration.getType("typ.langZeitZählstelle").getElements();

            // ... und jetzt Liste mit zu verwendenden aufbauen (falls nicht "alle" verwendet werden sollen) ...
            boolean isAlle = ((_listePidsLzzs.size() == 1) && _listePidsLzzs.get(0).equals("alle"));

            for (SystemObject so : listeAlleLzzsSystemObjekte) {
                if (isAlle || _listePidsLzzs.contains(so.getPid())) {
                    _listeLzzsSystemObjekte.add(so);
                }
            }

            // .. und für Konfigurationszusammenfassung Liste der nicht verwendeten PID's aus dem Aufrufparameter ermitteln
            List<String> listePidsVerwendet = new ArrayList<String>();

            for (SystemObject so : _listeLzzsSystemObjekte) {
                listePidsVerwendet.add(so.getPid());
            }

            for (String pid : _listePidsLzzs) {
                if (!listePidsVerwendet.contains(pid)) {
                    _listePidsLzzsNichtVorhanden.add(pid);
                }
            }
        }
        catch (Exception e) {
            debug.error("Es konnten keine Objekte vom typ.langZeitZählstelle ermittelt werden!. Programm wird beendet.", e);

            throw new RuntimeException();
        }

        if (_listeLzzsSystemObjekte.size() == 0) {
            debug.error("Es konnten keine Objekte vom typ.langZeitZählstelle ermittelt werden!. Programm wird beendet.");

            throw new RuntimeException();
        }
    }

    /**
     * Wertet die Aufrufparameter
     * <p>
     * -vonZeitPunkt
     * <p>
     * -intervallDauerInStunden
     * <p>
     * -nachberechnungInStunden
     * <p>
     * wie folgt aus:
     * <p>
     * <tt>-intervallDauerInStunden</tt>: Wird auf gültige Werte (1,2,3,4,6,8,12,24) geprüft und bei ungültigem Wert auf
     * 1 Stunde gesetzt.
     * <p>
     * <tt>-nachberechnungInStunden</tt>: Wird auf Gültigkeit getestet (ganze Zahl) und bei ungültigem Wert auf 0
     * Stunden (keine Nachberechnung) gesetzt.
     * <p>
     * <tt>-vonZeitPunkt</tt>: Wird auf gültigen Intervallanfangszeitpunkt abhängig vom Intervalltyp und
     * <tt>-nachberechnungInStunden</tt> auf denn nächsten, davorliegenden Intervallbeginn gesetzt.
     * <p>
     * Beispiel: Ist <tt>-vonZeitPunkt=04.11.2010 11:51:01,956</tt>, betrachtets Intervall 1 Stunde und
     * <tt>-nachberechnungInStunden=7</tt>, so wird der Zeitpunkt auf <tt>04.11.2010 04:00:00,000</tt> gesetzt.
     */
    private void auswertungAufrufParameterVonBisIntervallNachberechnung() {

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -nachberechnungInStunden
        // ------------------------------------------------------------------------------------------------------
        try {
            _nachberechnungInStunden = new Integer(_nachberechnungInStundenAufrufParameter);

            if (_nachberechnungInStunden < 0) {
                throw new NumberFormatException();
            }
        }
        catch (NumberFormatException e) {
            _nachberechnungInStunden = 0;
            debug.warning(String.format("Aufrufparameter [-nachberechnungInStunden=%s] ist ungültig. Wert wurde für weiter Berechnung auf [-nachberechnungInStunden=%d] gesetzt.", _nachberechnungInStundenAufrufParameter, _nachberechnungInStunden));
        }

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -intervallDauerInStunden
        // ------------------------------------------------------------------------------------------------------
        try {
            Set<Integer> gueltigeWerte = new HashSet<Integer>();

            gueltigeWerte.add(1);
            gueltigeWerte.add(2);
            gueltigeWerte.add(3);
            gueltigeWerte.add(4);
            gueltigeWerte.add(6);
            gueltigeWerte.add(8);
            gueltigeWerte.add(12);
            gueltigeWerte.add(24);
            _intervallDauerInStunden = new Integer(_intervallDauerInStundenAufrufParameter);

            if (!gueltigeWerte.contains(_intervallDauerInStunden)) {
                throw new NumberFormatException();
            }
        }
        catch (NumberFormatException e) {
            _intervallDauerInStunden = 1;
            debug.warning(String.format("Aufrufparameter [-intervallDauerInStunden=%s] ist ungültig (nur 1,2,3,4,6,8,12,24 zulässig). Wert wurde für weiter Berechnung auf [-intervallDauerInStunden=%d] gesetzt.", _intervallDauerInStundenAufrufParameter, _intervallDauerInStunden));
        }

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -vonZeitPunkt
        // ------------------------------------------------------------------------------------------------------
        _vonZeitPunkt = ermittleIntervallBeginn(_vonZeitPunktAufrufParameter, _intervallDauerInStunden);
        _bisZeitPunkt = ermittleIntervallEnde(_vonZeitPunktAufrufParameter, _intervallDauerInStunden);

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -minimaleWarteZeitInMinuten
        // ------------------------------------------------------------------------------------------------------
        try {
            _minimaleWarteZeitInMinuten = new Integer(_minimaleWarteZeitInMinutenAufrufParameter);

            if (_minimaleWarteZeitInMinuten <= 0) {
                throw new NumberFormatException();
            }
        }
        catch (NumberFormatException e) {
            _minimaleWarteZeitInMinuten = 5;
            debug.warning(String.format("Aufrufparameter [-minimaleWarteZeitInMinuten=%s] ist ungültig. Wert wurde für weiter Berechnung auf [-minimaleWarteZeitInMinuten=%d] gesetzt.", _minimaleWarteZeitInMinutenAufrufParameter, _minimaleWarteZeitInMinuten));
        }
    }

    /**
     * Ermittelt den IntervallBeginn ausgehend von einem Zeitpunk, der Intervalldauer und der Nachberechnungszeit
     *
     * @param spaetesterIntervallBeginn Anfangszeitpunkt in ms seit 1.1.1970, ab dem spätestens die Daten berechnet
     *                                  werden sollen. Dieser Wert wird dann noch unter Berücksichtigung der Parameter
     *                                  <tt>intervallDauerInStunden</tt>  soweit in die Vergangenheit verschoben, dass
     *                                  genau ein gültiger Intervallbeginn ermittelt wird.
     * @param intervallDauerInStunden   Intervalldauer, die gerechnet werden soll.
     *
     * @return Intervallbeginn (GENAU!!), der vor dem Zeitpunkt <tt>vonZeitPunkt</tt>
     */
    private long ermittleIntervallBeginn(long spaetesterIntervallBeginn, int intervallDauerInStunden) {
        Calendar vonZeitPunkt = new GregorianCalendar();

        // vonZeitPunkt darf spätestens sein...
        vonZeitPunkt.setTimeInMillis(spaetesterIntervallBeginn);

        // ... und muss jetzt auf gültigen Intervallzeitpunkt DAVOR abgebildet werden...
        // ToDo: Vereinfachte Version, liefert nur für Intervalldauer = 1h korrekte Werte, indem einfach die Minuten,
        switch (intervallDauerInStunden) {
        case 1 :

            // Sekunden und Millisekunden auf 0 gesetzt werden.
            vonZeitPunkt.set(MILLISECOND, 0);
            vonZeitPunkt.set(SECOND, 0);
            vonZeitPunkt.set(MINUTE, 0);

            break;

        default :
            debug.warning("IntervallDauer != 1 Stunde wird aktuell nicht unterstützt! Intervallbeginn wird für Intervalldauer == 1 Stunde berechnet.");

            // Sekunden und Millisekunden auf 0 gesetzt werden.
            vonZeitPunkt.set(MILLISECOND, 0);
            vonZeitPunkt.set(SECOND, 0);
            vonZeitPunkt.set(MINUTE, 0);
        }

        return vonZeitPunkt.getTimeInMillis();
    }

    /**
     * Ermittelt das IntervallEnde ausgehend von einem Zeitpunkt und der Intervalldauer
     *
     * @param fruehestesIntervallEnde Endzeitpunkt, bis zu dem mindestens gerechnet werden muss.
     * @param intervallDauerInStunden Intervalldauer, die gerechnet werden soll.
     *
     * @return Intervallende (AUSSCHLIESSLICH!!), also der Intervallbeginn des nächsten Intervalls
     */
    private long ermittleIntervallEnde(long fruehestesIntervallEnde, int intervallDauerInStunden) {
        return ermittleIntervallBeginn(fruehestesIntervallEnde + (intervallDauerInStunden * 60 * 60 * 1000), intervallDauerInStunden);
    }

    /**
     * Ermittelt aus allen FS und den zugeordneten DE die DELve255 für jeden FS, damit für diesen die Pufferabfrage
     * durchgeführt werden kann.
     *
     * @param mapDeLveZuDe255 Map mit der Zuordnung der DeLve zur DeLve255 für die Pufferabfrage
     * @param mapDeLveZuFs    Map mit der Zuordnung der DeLve zum FS einer LZZS
     *
     * @return Map mit den DeLve255 und der Menge der FS, deren Daten über diese De aus dem Puffer abgerufen werden.
     */
    private Map<SystemObject, Set<SystemObject>> ermittleZuordnungDe255ZuFs(Map<SystemObject, SystemObject> mapDeLveZuDe255, Map<SystemObject, SystemObject> mapDeLveZuFs) {
        Map<SystemObject, Set<SystemObject>> mapDe255ZuFs = new HashMap<SystemObject, Set<SystemObject>>();

        for (SystemObject soDe : mapDeLveZuFs.keySet()) {

            // Aus der anderen Map die zur De passende De255 holen
            SystemObject soDeLve255 = mapDeLveZuDe255.get(soDe);

            // Testen, ob für die DE255 bereits ein Eintrag in der Ergebnisliste vorliegt, sonst neu anlegen
            Set<SystemObject> mengeFs = mapDe255ZuFs.get(soDeLve255);

            if (mengeFs == null) {

                // Eintrag und Menge für FS existieren noch nicht
                mengeFs = new HashSet<SystemObject>();
            }

            // Menge existiert jetzt in jedem Fall, dann einfach den zugeordneten FS holen und eintragen
            mengeFs.add(mapDeLveZuFs.get(soDe));
            mapDe255ZuFs.put(soDeLve255, mengeFs);
        }

        return mapDe255ZuFs;
    }

    /**
     * Ermittelt die Liste der Zuordnungen der DeLve zur DeLv2255 am selben SM, also welche DeLve über welchen
     * Sammelkanal 255 abgerufen werden können.
     *
     * @return Map mit der Zuordnung der DeLve zur DeLve255 für die Pufferabfrage
     */
    private Map<SystemObject, SystemObject> ermittleZuordnungDeLveZuDe255() {
        Map<SystemObject, SystemObject> mapDeLveZuDe255 = new HashMap<SystemObject, SystemObject>();
        AttributeGroup                  atg             = _konfiguration.getAttributeGroup("atg.de");
        Aspect                          asp             = _konfiguration.getAspect("asp.eigenschaften");

        // Hole alle SM
        List<SystemObject> listeSm = _konfiguration.getType("typ.steuerModul").getElements();

        for (SystemObject soSm : listeSm) {

            // Hole alle EAK am SM
            List<SystemObject> listeEak = ((ConfigurationObject) soSm).getNonMutableSet("Eak").getElements();

            // Hole alle DE an den EAK
            List<SystemObject> listeAllerDeAmSm = new ArrayList<SystemObject>();

            for (SystemObject soEak : listeEak) {
                List<SystemObject> listeDe = ((ConfigurationObject) soEak).getNonMutableSet("De").getElements();

                listeAllerDeAmSm.addAll(listeDe);
            }

            // Ermittele das DE255 in der FG1, wenn es keins gibt, dann keine weiteren Eintragungen für diesen SM
            _konfiguration.getConfigurationData(listeAllerDeAmSm.toArray(new SystemObject[listeAllerDeAmSm.size()]), atg, asp);

            SystemObject soDeLve255 = null;

            for (SystemObject soDe : listeAllerDeAmSm) {
                Data[] daten = _konfiguration.getConfigurationData(new SystemObject[] { soDe }, atg, asp);

                if (soDe.isOfType("typ.deLve") && (daten[0].getUnscaledValue("DEKanal").intValue() == 255)) {
                    soDeLve255 = soDe;

                    break;
                }
            }

            if (soDeLve255 != null) {

                // Ermittle alle DE der FG1 und speichere Paar DE-LVE:DE-LVE-255 in der Map
                for (SystemObject soDe : listeAllerDeAmSm) {
                    Data[] daten = _konfiguration.getConfigurationData(new SystemObject[] { soDe }, atg, asp);

                    if (soDe.isOfType("typ.deLve") && (daten[0].getUnscaledValue("DEKanal").intValue() != 255)) {
                        mapDeLveZuDe255.put(soDe, soDeLve255);
                    }
                }
            }
        }

        return mapDeLveZuDe255;
    }

    /**
     * Ermittelt zu allen LZZS die DE an den FS der MQ der AWQ der LZZS.
     *
     * @param listeLzzsSystemObjekte Liste mit allen LZZS, für die die Zuordnung DE-FS ermeittelt werden soll.
     *
     * @return Map mit der Zuordung DE-FS für die LZZS.
     */
    private Map<SystemObject, SystemObject> ermittleZuordnungDeZuFsAusLzzs(List<SystemObject> listeLzzsSystemObjekte) {
        Map<SystemObject, SystemObject> mapDeLveZuFs = new HashMap<SystemObject, SystemObject>();

        for (SystemObject lzzs : listeLzzsSystemObjekte) {

            // ------------------------------------------------------------------------------------------------
            // MQ inRichung und gegenRichtung des LZZS ermitteln
            // ------------------------------------------------------------------------------------------------
            SystemObject soAwqInRichtung, soAwqGegenRichtung, soMqInRichtung, soMqGegenRichtung, soDe;

            // ATG an der LZZS mit den Referenzen auf die AWQ ermitteln
            AttributeGroup atg   = _konfiguration.getAttributeGroup("atg.langZeitZählstelle");
            Aspect         asp   = _konfiguration.getAspect("asp.eigenschaften");
            Data[]         daten = _konfiguration.getConfigurationData(new SystemObject[] { lzzs }, atg, asp);

            // AWQ InRichung ermitteln
            soAwqInRichtung = daten[0].getItem("InRichtung").getReferenceValue("Auswertequerschnitt").getSystemObject();

            // AWQ GegenRichung ermitteln
            soAwqGegenRichtung = daten[0].getItem("GegenRichtung").getReferenceValue("Auswertequerschnitt").getSystemObject();

            // MQ InRichtung ermitteln
            atg            = _konfiguration.getAttributeGroup("atg.auswerteQuerschnitt");
            asp            = _konfiguration.getAspect("asp.eigenschaften");
            daten          = _konfiguration.getConfigurationData(new SystemObject[] { soAwqInRichtung }, atg, asp);
            soMqInRichtung = daten[0].getReferenceValue("Messquerschnitt").getSystemObject();

            // MQ GegenRichtung ermitteln
            daten             = _konfiguration.getConfigurationData(new SystemObject[] { soAwqGegenRichtung }, atg, asp);
            soMqGegenRichtung = daten[0].getReferenceValue("Messquerschnitt").getSystemObject();

            // ATG mir FS-Informationen und DE-Referenz initialisieren
            atg = _konfiguration.getAttributeGroup("atg.fahrStreifen");
            asp = _konfiguration.getAspect("asp.eigenschaften");

            // Menge mit den FS am MQ InRichtung ermitteln
            List<SystemObject> listeSoFS = ((ConfigurationObject) soMqInRichtung).getNonMutableSet("FahrStreifen").getElements();

            for (SystemObject soFs : listeSoFS) {

                // Referenzierte DE ermitteln und Paar DE-FS in der Map speichern
                daten = _konfiguration.getConfigurationData(new SystemObject[] { soFs }, atg, asp);
                soDe  = daten[0].getReferenceValue("FahrStreifenQuelle").getSystemObject();

                if (soDe != null) {
                    mapDeLveZuFs.put(soDe, soFs);
                }
            }

            // Menge mit den FS am MQ GegenRichtung ermitteln
            listeSoFS = ((ConfigurationObject) soMqGegenRichtung).getNonMutableSet("FahrStreifen").getElements();

            for (SystemObject soFs : listeSoFS) {

                // Referenzierte DE ermitteln und Paar DE-FS in der Map speichern
                daten = _konfiguration.getConfigurationData(new SystemObject[] { soFs }, atg, asp);
                soDe  = daten[0].getReferenceValue("FahrStreifenQuelle").getSystemObject();

                if (soDe != null) {
                    mapDeLveZuFs.put(soDe, soFs);
                }
            }
        }

        return mapDeLveZuFs;
    }

    /**
     * Folgeauftrag für den aktuell übergebenen Auftrag berechnen.
     * <p>
     * Der Folgeauftrag liegt genau eine Intervalllänge später.
     *
     * @param auftrag Der aktuelle Auftrag, für den ein Folgeauftrag berechnet werden soll.
     *
     * @return Folgeauftrag für den aktuell als Parameter übergebenen Auftrag oder <code>null</code>, wenn kein
     *         Folgeauftrag mehr existiert.
     */
    private Auftrag erstelleFolgeAuftrag(Auftrag auftrag) {
        return new Auftrag(auftrag);
    }

    /** Initialisierung von PufferAbrufLzzs */
    private void initialisierung() {

        // Map aller DE-LVE und der zugeordneten DE255 der LVE ermitteln
        _mapDeLveZuDe255 = ermittleZuordnungDeLveZuDe255();

        // Aus den LZZS alle FS und die zugeordneten DE-LVE ermitteln
        _mapDeLveZuFs = ermittleZuordnungDeZuFsAusLzzs(_listeLzzsSystemObjekte);

        // Über die DE-LVE die Liste mit den DE255 der LVE und den zugeordneten FS ermitteln,
        // aus dieser Liste dann die eigentlichen Verwaltungsobjekte VewDeLve255 konstruieren.
        _mapDe255ZuFs = ermittleZuordnungDe255ZuFs(_mapDeLveZuDe255, _mapDeLveZuFs);

        // VewDeLve255-Objekte konstruieren
        for (SystemObject deLve255 : _mapDe255ZuFs.keySet()) {
            VewDeLve255 vewDeLve255 = new VewDeLve255(_dav, deLve255, _mapDe255ZuFs.get(deLve255), _intervallDauerInStunden, _nachberechnungInStunden);

            _mapVewDeLve255.put(deLve255.getPid(), vewDeLve255);
        }

        // Initiales setzen der Auftragsliste auf Basis der VewDe255Lve und der Aurufparameter
        for (VewDeLve255 vewDeLve255 : _mapVewDeLve255.values()) {
            long          intervallBeginn       = _vonZeitPunkt;
            long          intervallEnde         = _bisZeitPunkt;
            long          ausfuehrungsZeitpunkt = intervallEnde + (_minimaleWarteZeitInMinuten * 60 * 1000);
            Auftrag       auftrag               = new Auftrag(ausfuehrungsZeitpunkt, intervallBeginn, intervallEnde, vewDeLve255);
            AuftragsListe auftragsListe         = AuftragsListe.getInstanz();

            auftragsListe.addAuftrag(auftrag);
        }
    }

    /**
     * Initialisierung. Wird durch den {@link de.bsvrz.sys.funclib.application.StandardApplicationRunner} aufgerufen.
     *
     * @param dav Die Datenverteilerverbindung.
     */
    @Override
    public void initialize(ClientDavInterface dav) {
        _dav           = dav;
        _konfiguration = _dav.getDataModel();
    }

    /**
     * Wertet die Aufrufparameter aus.
     *
     * @param argumentList Aufrufparamter der Applikation
     *
     * @throws Exception Wenn die Aufrufparameter nicht vollständig ausgewertet werden können.
     */
    @Override
    public void parseArguments(ArgumentList argumentList) throws Exception {
        debug = Debug.getLogger();

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -listePids
        // ------------------------------------------------------------------------------------------------------
        // _listePidsAufrufParameter = argumentList.fetchArgument("-listePids=dz.scheppau.a39.49_400").asNonEmptyString();
        //
        _listePidsAufrufParameter = argumentList.fetchArgument("-listePids=alle").asNonEmptyString();

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -vonZeitPunkt
        // ------------------------------------------------------------------------------------------------------
        try {
            _vonZeitPunktAufrufParameter = argumentList.fetchArgument("-vonZeitPunkt=jetzt").asTime();
            _vonZeitPunktAufrufParameter = argumentList.fetchArgument("-vonZeitPunkt=08.12.2010 05:00:01,123").asTime();
        }
        catch (Exception e) {
            debug.error(String.format("Ungültiges Argument beim Aufrufparameter [-vonZeitPunkt]. Programm wird beendet.%nFehler: %s", e));

            throw new RuntimeException(e);
        }

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -intervallDauerInStunden
        // ------------------------------------------------------------------------------------------------------
        _intervallDauerInStundenAufrufParameter = argumentList.fetchArgument("-intervallDauerInStunden=1").asNonEmptyString().trim();

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -nachberechnungInStunden
        // ------------------------------------------------------------------------------------------------------
        _nachberechnungInStundenAufrufParameter = argumentList.fetchArgument("-nachberechnungInStunden=24").asNonEmptyString().trim();

        // ------------------------------------------------------------------------------------------------------
        // Aufrufparameter: -minimaleWarteZeitInMinuten
        // ------------------------------------------------------------------------------------------------------
        _minimaleWarteZeitInMinutenAufrufParameter = argumentList.fetchArgument("-minimaleWarteZeitInMinuten=5").asNonEmptyString().trim();
    }

    /** Startet die eigentliche Aggregation, werte die Auftragsliste aus und aggregiert die Daten. */
    private void run() {
        debug.info(String.format("%n=================================================================%n" + "Programm Aggregation Langzeitdaten DTV gestartet...%n" + "=================================================================%n"));

        // Auftragsliste holen
        AuftragsListe auftragsListe = AuftragsListe.getInstanz();

        // Endlosschleife mit Prüfung der Auftragsliste.
        // Sobald keinerlei Aufträge mehr vorhanden sind wird das Programm beendet
        while (auftragsListe.hatNochAuftraege()) {
            Auftrag auftrag;
            Auftrag folgeAuftrag;

            if ((auftrag = auftragsListe.holeNaechstenAuftrag()) != null) {
                StringBuffer statusMeldung = new StringBuffer();

                // Ggf. Folgeauftrag in Liste eintragen
                folgeAuftrag = erstelleFolgeAuftrag(auftrag);
                statusMeldung.append(String.format("Aktueller Auftrag: %s", auftrag));

                if (folgeAuftrag != null) {
                    auftragsListe.addAuftrag(folgeAuftrag);
                    statusMeldung.append(String.format("%nFolgeauftrag     : %s", folgeAuftrag));
                } else {
                    statusMeldung.append(String.format("%nFolgeauftrag     : %s", "kein weiterer Folgeauftrag mehr zu berücksichtigen"));
                }

                debug.info(statusMeldung.toString());
                auftrag.ausfuehren();
                debug.info(String.format("Ausgeführter Auftrag: %s", auftrag));
            } else {

                // Zurzeit ist nichts zu tun, also warten und es dann noch mal versuchen...
                debug.info(String.format("Zur Zeit keine sofort auszuführenden Aufträge in der Auftragsliste. Warte %d Sekunden.", WARTEZEIT_AUF_NAECHSTEN_AUFTRAG / 1000));

                try {
                    Thread.sleep(WARTEZEIT_AUF_NAECHSTEN_AUFTRAG);
                }
                catch (InterruptedException e) {
                    debug.info("Sleep der Auftragsbearbeitung wurde vorzeitig unterbrochen: ", e);
                }
            }
        }

        debug.info(String.format("%n=================================================================%n" + "Programm PufferAbruf Langzeitdaten wird ordnungsgemäß beendet,%n" + "da keinerlei Aufträge (weder aktuell noch zukünftig) anliegen%n" + "=================================================================%n"));
        System.exit(0);
    }
}


//~Formatiert mit 'inovat Kodierkonvention' am 08.12.10
